@@ -5,7 +5,7 @@
 extern "C" {
 #include "fts-xapian-plugin.h"
 }
-#include <ftw.h>
+#include <dirent.h>
 #include <unicode/unistr.h>
 #include <sys/time.h>
 
@@ -14,8 +14,8 @@ extern "C" {
 #define XAPIAN_COMMIT_ENTRIES 1000000L
 #define XAPIAN_COMMIT_TIMEOUT 300L
 #define XAPIAN_WILDCARD "wldcrd"
-#define XAPIAN_EXPUNGE_SIZE 3
 #define XAPIAN_EXPUNGE_HEADER 9
+#define XAPIAN_MIN_RAM 200L
 
 #define HDRS_NB 11
 static const char * hdrs_emails[HDRS_NB] = { "uid", "subject", "from", "to",  "cc",  "bcc",  "messageid", "listid", "body", "expungeheader",	""  };
@@ -28,8 +28,6 @@ struct xapian_fts_backend
         struct fts_backend backend;
         char * path;
 	long partial,full;
-	bool attachments;
-	bool doexpunge;
 
 	char * guid;
 	char * boxname;
@@ -47,12 +45,16 @@ struct xapian_fts_backend
 	long perf_nb;
 	long perf_uid;
 	long perf_dt;
+
+	long nb_pushes;
+	long max_push;
 };
 
 struct xapian_fts_backend_update_context
 {
         struct fts_backend_update_context ctx;
         char * tbi_field=NULL;
+	bool isattachment=false;
         bool tbi_isfield;
         uint32_t tbi_uid=0;
 };
@@ -85,12 +87,13 @@ static int fts_backend_xapian_init(struct fts_backend *_backend, const char **er
 	backend->path = NULL;
 	backend->old_guid = NULL;
 	backend->old_boxname = NULL;
-	backend->attachments = false;
-	backend->doexpunge = false;
 	verbose = 0;
 	backend->partial = 0;
 	backend->full = 0;
 
+	backend->nb_pushes=0;
+	backend->max_push=0;
+
 	env = mail_user_plugin_getenv(_backend->ns->user, "fts_xapian");
 	if (env == NULL) 
 	{
@@ -117,7 +120,7 @@ static int fts_backend_xapian_init(struct fts_backend *_backend, const char **er
 		}
 		else if (strncmp(*tmp,"attachments=",12)==0)
 		{
-			if(atol(*tmp + 12)>0) backend->attachments=true;
+			// Legacy
 		}
 		else 
 		{
@@ -160,7 +163,7 @@ static int fts_backend_xapian_init(struct fts_backend *_backend, const char **er
 		}
 	}
 
-	if(verbose>0) i_info("FTS Xapian: Starting with partial=%ld full=%ld attachments=%d verbose=%d",backend->partial,backend->full,backend->attachments,verbose);
+	if(verbose>0) i_info("FTS Xapian: Starting with partial=%ld full=%ld verbose=%d",backend->partial,backend->full,verbose);
 
 	return 0;
 }
@@ -350,8 +353,6 @@ static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_co
 		return FALSE;
 	}
 
-	backend->doexpunge = true;
-
 	if((backend->old_guid == NULL) || (strcmp(backend->old_guid,backend->guid)!=0))
         {
 		fts_backend_xapian_oldbox(backend);
@@ -386,20 +387,28 @@ static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_co
 	const char * type = key->body_content_type;
 	const char * disposition = key->body_content_disposition;
 
-	if(verbose>1) i_info("FTS Xapian: New part (Header=%s,Type=%s,Disposition=%s)",field,type,disposition);
+	if(verbose>0) i_info("FTS Xapian: New part (Header=%s,Type=%s,Disposition=%s)",field,type,disposition);
 
 	// Verify content-type
-	if((type != NULL) && (strncmp(type,"text",4)!=0))
+
+	if(key->type == FTS_BACKEND_BUILD_KEY_BODY_PART_BINARY)
 	{
-		if(verbose>0) i_info("FTS Xapian: Skipping part of type '%s'",type);
+		if(verbose>0) i_info("FTS Xapian: Skipping binary part of type '%s'",type);
+		return FALSE;
+	}
+
+	if((type != NULL) && (strncmp(type,"text",4)!=0) && ((disposition==NULL) || ((strstr(disposition,"filename=")==NULL) && (strstr(disposition,"attachment")==NULL))))
+	{
+		if(verbose>0) i_info("FTS Xapian: Non-binary & non-text part of type '%s'",type);
 		return FALSE;
 	}
 
 	// Verify content-disposition
-	if((disposition != NULL) && (!backend->attachments) && ((strstr(disposition,"filename=")!=NULL) || (strstr(disposition,"attachment")!=NULL)))
+	ctx->isattachment=false;
+	if((disposition != NULL) && ((strstr(disposition,"filename=")!=NULL) || (strstr(disposition,"attachment")!=NULL)))
 	{
-		if(verbose>0) i_info("FTS Xapian: Skipping part of type '%s' and disposition '%s'",type,disposition);
-		return FALSE;
+		if(verbose>0) i_info("FTS Xapian: Found part as attachment of type '%s' and disposition '%s'",type,disposition);
+		ctx->isattachment=true;		
 	}
 
 	// Fill-in field
@@ -427,7 +436,7 @@ static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_co
 	}
 	if(i>=HDRS_NB) 
 	{ 
-		if(verbose>1) i_info("FTS Xapian: Unknown header (indexing) '%s'",ctx->tbi_field); 
+		if(verbose>0) i_info("FTS Xapian: Unknown header '%s' of part",ctx->tbi_field); 
 		i_free(ctx->tbi_field); 
 		ctx->tbi_field=NULL;
 		return FALSE; 
@@ -483,12 +492,23 @@ static int fts_backend_xapian_refresh(struct fts_backend * _backend)
 
 static int fts_backend_xapian_update_build_more(struct fts_backend_update_context *_ctx, const unsigned char *data, size_t size)
 {
-	if(verbose>1) i_info("FTS Xapian: fts_backend_xapian_update_build_more");
-
-	struct xapian_fts_backend_update_context *ctx =
-		(struct xapian_fts_backend_update_context *)_ctx;
-	struct xapian_fts_backend *backend =
-                (struct xapian_fts_backend *) ctx->ctx.backend;
+	struct xapian_fts_backend_update_context *ctx = (struct xapian_fts_backend_update_context *)_ctx;
+	struct xapian_fts_backend *backend = (struct xapian_fts_backend *) ctx->ctx.backend;
+	
+	if(verbose>0)
+	{
+		if(ctx->isattachment) 
+		{
+			char * t = i_strdup("NODATA");
+			if(data != NULL) { i_free(t); t = i_strndup(data,40); } 
+			i_info("FTS Xapian: Indexing part as attachment (data like '%s')",t);
+			i_free(t);
+		}
+		else
+		{
+			i_info("FTS Xapian: Indexing part as text");
+		}
+	}
 
 	if(ctx->tbi_uid<1) return 0;
 
@@ -508,8 +528,9 @@ static int fts_backend_xapian_update_build_more(struct fts_backend_update_contex
 	gettimeofday(&tp, NULL);
 	long current_time = tp.tv_sec * 1000 + tp.tv_usec / 1000;
 	
-	if(!fts_backend_xapian_test_memory())
+	if(!fts_backend_xapian_test_memory(backend))
 	{
+		if(verbose>0) i_info("FTS Xapian: Warning Low memory");
 		fts_backend_xapian_release(backend,"Low memory indexing", current_time);
 		if(!fts_backend_xapian_check_access(backend))
 		{
@@ -546,18 +567,38 @@ static int fts_backend_xapian_update_build_more(struct fts_backend_update_contex
 
 static int fts_backend_xapian_optimize(struct fts_backend *_backend)
 {
-	if(verbose>0) i_info("FTS Xapian: fts_backend_xapian_optimize");
-
 	struct xapian_fts_backend *backend =
 		(struct xapian_fts_backend *) _backend;
 
-	struct timeval tp;
-        gettimeofday(&tp, NULL);
-        long current_time = tp.tv_sec * 1000 + tp.tv_usec / 1000;
-	
-	fts_backend_xapian_do_expunge(backend,"optimize",0);
-        fts_backend_xapian_release(backend,"optimize",current_time);
+	i_info("FTS Xapian: fts_backend_xapian_optimize '%s'",backend->path);
 
+	struct stat sb;
+	if(!( (stat(backend->path, &sb)==0) && S_ISDIR(sb.st_mode)))
+	{
+		i_error("FTS Xapian: Index folder inexistent");
+		return -1;
+	}
+
+	DIR* dirp = opendir(backend->path);
+	struct dirent * dp;
+	char *s;
+    	while ((dp = readdir(dirp)) != NULL)
+	{
+		s = i_strdup_printf("%s/%s",backend->path,dp->d_name);
+	
+		if((dp->d_type == DT_REG) && (strncmp(dp->d_name,"expunge_",8)==0))
+		{
+			i_info("Removing %s",s);
+			remove(s);
+		}
+		else if((dp->d_type == DT_DIR) && (strncmp(dp->d_name,"db_",3)==0))
+		{
+			i_info("Expunging %s",s);
+			fts_backend_xapian_do_expunge(s);
+		}
+		i_free(s);
+	}
+	closedir(dirp);
 	return 0;
 }
 
@@ -575,7 +616,41 @@ static int fts_backend_xapian_rescan(struct fts_backend *_backend)
 		return -1;
 	}
 
-	return ftw(backend->path,fts_backend_xapian_empty_db,512);
+	DIR* dirp = opendir(backend->path);
+	struct dirent * dp;
+	char *s,*s2;
+	while ((dp = readdir(dirp)) != NULL)
+	{
+		s = i_strdup_printf("%s/%s",backend->path,dp->d_name);
+
+		if((dp->d_type == DT_REG) && (strncmp(dp->d_name,"expunge_",8)==0))
+		{
+			i_info("Removing[1] %s",s);
+			remove(s);
+		}
+		else if((dp->d_type == DT_DIR) && (strncmp(dp->d_name,"db_",3)==0))
+		{
+			DIR * d2 = opendir(s);
+			struct dirent *dp2;
+			while ((dp2 = readdir(d2)) != NULL)
+			{
+				s2 = i_strdup_printf("%s/%s",s,dp2->d_name);
+				if(dp2->d_type == DT_REG)
+				{
+					i_info("Removing[2] %s",s2);
+					remove(s2);
+				}
+				i_free(s2);
+			}
+			closedir(d2);
+			i_info("Removing dir %s",s);
+			remove(s);
+		}
+		i_free(s);
+	}
+	closedir(dirp);
+
+	return 0;
 }
 
 static int fts_backend_xapian_lookup(struct fts_backend *_backend, struct mailbox *box, struct mail_search_arg *args, enum fts_lookup_flags flags, struct fts_result *result)
@@ -622,6 +697,7 @@ static int fts_backend_xapian_lookup(struct fts_backend *_backend, struct mailbo
 	XResultSet * r=fts_backend_xapian_query(dbr,qs);
 
 	long n=r->size;
+	if(verbose>0) { i_info("FTS Xapian: QUery '%s' -> %ld results",qs->get_string().c_str(),n); }
 
 	i_array_init(&(result->definite_uids),r->size);
 
@@ -654,7 +730,7 @@ static int fts_backend_xapian_lookup(struct fts_backend *_backend, struct mailbo
 	return 0;
 }
 
-static int fts_backend_xapian_lookup_multi (struct fts_backend *_backend, struct mailbox *const boxes[], struct mail_search_arg *args, enum fts_lookup_flags flags, struct fts_multi_result *result)
+static int fts_backend_xapian_lookup_multi(struct fts_backend *_backend, struct mailbox *const boxes[], struct mail_search_arg *args, enum fts_lookup_flags flags, struct fts_multi_result *result)
 {
 	if(verbose>0) i_info("FTS Xapian: fts_backend_xapian_lookup_multi");
 
@@ -662,6 +738,7 @@ static int fts_backend_xapian_lookup_multi (struct fts_backend *_backend, struct
                 (struct xapian_fts_backend *) _backend;
 
 	ARRAY(struct fts_result) box_results;
+
 	struct fts_result *box_result;
 	int i;
 
@@ -670,8 +747,17 @@ static int fts_backend_xapian_lookup_multi (struct fts_backend *_backend, struct
 	{
 		box_result = array_append_space(&box_results);
 		box_result->box = boxes[i];
-		if(fts_backend_xapian_lookup(_backend, boxes[i], args, flags, box_result)<1) return -1;
+		if(fts_backend_xapian_lookup(_backend, boxes[i], args, flags, box_result)<0) 
+		{
+			void* p=&box_results;
+			p_free(result->pool, p);
+			return -1;
+		}
 	}
+
+	array_append_zero(&box_results);
+	result->box_results = array_idx_modifiable(&box_results, 0);
+
 	return 0;
 }
 
